import logging

from boto3 import exceptions as boto3_exceptions
from botocore import stub
from tornado import testing
import botocore.exceptions
import mock

from kingpin.actors import exceptions
from kingpin.actors.aws import base
from kingpin.actors.aws import settings
import importlib

log = logging.getLogger(__name__)

# STATIC VALUES FOR TESTS
TARGET_GROUP_RESPONSE = {
    "ResponseMetadata": {
        "HTTPHeaders": {
            "content-length": "1228",
            "content-type": "text/xml",
            "date": "Tue, 13 Feb 2018 17:50:56 GMT",
            "x-amzn-requestid": "123-123-123-123-123",
        },
        "HTTPStatusCode": 200,
        "RequestId": "123-123-123-123-123",
        "RetryAttempts": 0,
    },
    "TargetGroups": [
        {
            "HealthCheckIntervalSeconds": 30,
            "HealthCheckPath": "/",
            "HealthCheckPort": "traffic-port",
            "HealthCheckProtocol": "HTTP",
            "HealthCheckTimeoutSeconds": 5,
            "HealthyThresholdCount": 5,
            "LoadBalancerArns": [],
            "Matcher": {"HttpCode": "200"},
            "Port": 80,
            "Protocol": "HTTP",
            "TargetGroupArn": "arn:aws:elb:us-east-1:123:targetgroup/unittest/123",
            "TargetGroupName": "kingpin-integration-test",
            "UnhealthyThresholdCount": 2,
            "VpcId": "vpc-123",
        }
    ],
}


class TestBase(testing.AsyncTestCase):
    def setUp(self):
        super(TestBase, self).setUp()
        settings.AWS_ACCESS_KEY_ID = "unit-test"
        settings.AWS_SECRET_ACCESS_KEY = "unit-test"
        settings.AWS_SESSION_TOKEN = "unit-test"
        importlib.reload(base)

    @testing.gen_test
    def test_api_call_raises_client_exception(self):
        actor = base.AWSBaseActor("Unit Test Action", {})
        stubber = stub.Stubber(actor.iam_conn)
        stubber.add_client_error("list_roles", 400, "Error")
        stubber.activate()
        with self.assertRaises(botocore.exceptions.ClientError):
            yield actor.api_call(actor.iam_conn.list_roles)
        stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_api_call_raises_boto3_core_exception(self):
        actor = base.AWSBaseActor("Unit Test Action", {})
        actor.iam_conn = mock.MagicMock()
        actor.iam_conn.list_roles.side_effect = boto3_exceptions.Boto3Error
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield actor.api_call(actor.iam_conn.list_roles)

    @testing.gen_test
    def test_parse_policy_json(self):
        actor = base.AWSBaseActor("Unit Test Action", {})

        # Should work fine by default with good data
        ret = actor._parse_policy_json("examples/aws.iam.user/s3_example.json")
        self.assertEqual(ret["Version"], "2012-10-17")

        # If the file doesn't exist, raise an exception
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            actor._parse_policy_json("junk")

    @testing.gen_test
    def test_parse_policy_json_none(self):
        actor = base.AWSBaseActor("Unit Test Action", {})
        ret = actor._parse_policy_json(None)
        self.assertEqual(ret, None)



import logging
import importlib
import json

from datetime import datetime

import mock

from botocore.stub import Stubber
from tornado import gen, testing

from kingpin.actors import exceptions
from kingpin.actors.aws import iam, settings

log = logging.getLogger(__name__)


@gen.coroutine
def tornado_value(*args):
    """Returns whatever is passed in. Used for testing."""
    raise gen.Return(*args)


class TestIAMBaseActor(testing.AsyncTestCase):
    def setUp(self):
        super(TestIAMBaseActor, self).setUp()
        settings.AWS_ACCESS_KEY_ID = "unit-test"
        settings.AWS_SECRET_ACCESS_KEY = "unit-test"
        settings.AWS_SESSION_TOKEN = "unit-test"
        importlib.reload(iam)

        # Create our actor object with some basics... then mock out the IAM
        # connections..
        self.actor = iam.IAMBaseActor(
            "Unit Test",
            {
                "name": "test",
                "state": "present",
                "inline_policies": "examples/aws.iam.user/s3_example.json",
            },
        )
        self.iam_stubber = Stubber(self.actor.iam_conn)

        # The base class defines these as None -- but in order to test them, we
        # need to have more realistic and trackable method names.
        self.actor.entity_name = "User"
        self.actor.create_entity = self.actor.iam_conn.create_user
        self.actor.delete_entity = self.actor.iam_conn.delete_user
        self.actor.delete_entity_policy = self.actor.iam_conn.delete_user_policy
        self.actor.get_entity = self.actor.iam_conn.get_user
        self.actor.list_entity_policies = self.actor.iam_conn.list_user_policies
        self.actor.get_entity_policy = self.actor.iam_conn.get_user_policy
        self.actor.put_entity_policy = self.actor.iam_conn.put_user_policy

        # Pretend like we're a more full featured actor (User/Group/Role) that
        # has inline policies. This could be abstracted differently by making
        # another BaseActor for Users/Groups/Roles thats separate from
        # InstanceProfiles -- but for now this will do.
        self.actor._parse_inline_policies(self.actor.option("inline_policies"))

    @testing.gen_test
    def test_generate_policy_name(self):
        name = "/some-?funky*-directory/with.my.policy.json"
        parsed = self.actor._generate_policy_name(name)
        self.assertEqual(parsed, "some-funky-directory-with.my.policy")

    @testing.gen_test
    def test_get_entity_policies_500(self):
        self.iam_stubber = Stubber(self.actor.iam_conn)
        self.iam_stubber.add_client_error("list_user_policies", "500", "Server Error!")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._get_entity_policies("test")

    @testing.gen_test
    def test_get_entity_policies_400(self):
        self.iam_stubber.add_client_error("list_user_policies", "400", "NoSuchEntity")
        self.iam_stubber.activate()
        ret = yield self.actor._get_entity_policies("test")
        self.assertEqual(ret, {})

    @testing.gen_test
    def test_get_entities_other_500(self):
        self.iam_stubber.add_response(
            # API Call
            "list_user_policies",
            # Response
            # Use one policy here so we do not have to deal with extra calls
            # and multiple Stubber responses.
            {"PolicyNames": ["test1"]},
            # Call Params
            {"UserName": "test"},
        )
        self.iam_stubber.add_client_error("get_user_policy", "500", "SomeError")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._get_entity_policies("test")

    @testing.gen_test
    def test_get_entity_policies(self):
        policy_str = "".join(
            [
                "%7B%22Version%22%3A%20%222012-10-17%22%2C%20",
                "%22Statement%22%3A%20%5B%7B%22Action%22%3A%20%5B",
                "%22s3%3ACreate%2A%22%2C%20%22s3%3AGet%2A%22%2C%20",
                "%22s3%3APut%2A%22%2C%20%22s3%3AList%2A%22%5D%2C%20",
                "%22Resource%22%3A%20%5B",
                "%22arn%3Aaws%3As3%3A%3A%3Akingpin%2A%2F%2A%22%2C%20",
                "%22arn%3Aaws%3As3%3A%3A%3Akingpin%2A%22%5D%2C%20",
                "%22Effect%22%3A%20%22Allow%22%7D%5D%7D",
            ]
        )
        policy_dict = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": ["s3:Create*", "s3:Get*", "s3:Put*", "s3:List*"],
                    "Resource": ["arn:aws:s3:::kingpin*/*", "arn:aws:s3:::kingpin*"],
                    "Effect": "Allow",
                }
            ],
        }

        # Return a list of entity policy names...
        fake_pols = ["test1", "test2", "test3"]
        self.iam_stubber.add_response(
            # API Call
            "list_user_policies",
            # Response
            {"PolicyNames": fake_pols},
            # Call Params
            {"UserName": "test"},
        )

        for pol in fake_pols:
            self.iam_stubber.add_response(
                # API Call
                "get_user_policy",
                # Response
                {"UserName": "test", "PolicyName": pol, "PolicyDocument": policy_str},
                # Call Params
                # Disable expected parameters because calls to
                # api_call(self.get_entity_policy, [...]) can come in out of order but boto3 Subber
                # expects them to be in order if you use this parameter. We do our own checks below
                # to ensure we have all the data we expected.
                # {"UserName": "test", "PolicyName": pol},
            )

        # Finally, make the call and see if we get all the policies
        self.iam_stubber.activate()
        ret = yield self.actor._get_entity_policies("test")
        self.assertEqual(len(ret), 3)
        self.assertEqual(ret["test1"], policy_dict)
        self.assertEqual(ret["test2"], policy_dict)
        self.assertEqual(ret["test3"], policy_dict)

    @testing.gen_test
    def test_parse_inline_policies(self):
        parsed_policy = self.actor.inline_policies["examples-aws.iam.user-s3_example"]
        self.assertEqual(parsed_policy["Version"], "2012-10-17")

    @testing.gen_test
    def test_parse_inline_policies_none(self):
        self.actor._parse_inline_policies(None)
        self.assertEqual(self.actor.inline_policies, None)

    @testing.gen_test
    def test_ensure_inline_policies(self):
        # First, pretend like there are a few policies in place and we're not
        # passing any in, however we are purging policies we don't manage.
        fake_pol = {
            "Policy1": {"junk": "policy"},
            "Policy2": {"more": "junk"},
        }
        self.actor._get_entity_policies = mock.MagicMock()
        self.actor._get_entity_policies.side_effect = [tornado_value(fake_pol)]

        # Mock out the delete_entity_policy and put_entity_policy methods
        self.actor._delete_entity_policy = mock.MagicMock()
        self.actor._delete_entity_policy.side_effect = [
            tornado_value(None),
            tornado_value(None),
        ]
        self.actor._put_entity_policy = mock.MagicMock()
        self.actor._put_entity_policy.side_effect = [tornado_value(None)]

        # Ensure that the new policy was pushed, and the old policies were
        # deleted
        yield self.actor._ensure_inline_policies("test")
        self.assertEqual(1, self.actor._put_entity_policy.call_count)
        self.actor._delete_entity_policy.assert_has_calls(
            [
                mock.call("test", "Policy1"),
                mock.call("test", "Policy2"),
            ],
            any_order=True,
        )

    @testing.gen_test
    def test_ensure_inline_policies_updated(self):
        # First, pretend like there are a few policies in place and we're not
        # passing any in, however we are purging policies we don't manage.
        fake_pol = {
            "Policy1": {"junk": "policy"},
            "examples-aws.iam.user-s3_example": {"more": "junk"},
        }
        self.actor._get_entity_policies = mock.MagicMock()
        self.actor._get_entity_policies.side_effect = [tornado_value(fake_pol)]
        self.actor._put_entity_policy = mock.MagicMock()
        self.actor._put_entity_policy.side_effect = [tornado_value(None)]

        self.iam_stubber.add_response(
            # API Call
            "delete_user_policy",
            # Response
            {},
            # Call Params
            {"UserName": "test", "PolicyName": "Policy1"},
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_inline_policies("test")
        self.assertEqual(1, self.actor._put_entity_policy.call_count)

    @testing.gen_test
    def test_delete_entity_policy_dry(self):
        self.actor._dry = True
        self.iam_stubber.activate()
        yield self.actor._delete_entity_policy("test", "test-policy")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_delete_entity_policy(self):
        self.iam_stubber.add_response(
            # API Call
            "delete_user_policy",
            # Response
            {},
            # Call Params
            {"UserName": "test", "PolicyName": "test-policy"},
        )
        self.iam_stubber.activate()
        yield self.actor._delete_entity_policy("test", "test-policy")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_delete_entity_policy_exception(self):
        self.iam_stubber.activate()
        self.iam_stubber.add_client_error("delete_user_policy", 500, "Yikes!")
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._delete_entity_policy("test", "test-policy")

    @testing.gen_test
    def test_put_entity_policy_dry(self):
        self.actor._dry = True
        self.iam_stubber.activate()
        yield self.actor._put_entity_policy("test", "test-policy", {})
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_put_entity_policy(self):
        self.iam_stubber.add_response(
            # API CALL
            "put_user_policy",
            # Response
            {},
            # Call Params
            {"UserName": "test", "PolicyName": "test-policy", "PolicyDocument": "{}"},
        )
        self.iam_stubber.activate()
        yield self.actor._put_entity_policy("test", "test-policy", {})
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_put_entity_policy_exception(self):
        self.iam_stubber.add_client_error("put_user_policy", 500, "Yikes!")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._put_entity_policy("test", "test-policy", {})

    @testing.gen_test
    def test_get_entity_500(self):
        # Test a random unexpected failure
        self.iam_stubber.add_client_error("get_user", 500, "Yikes!")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._get_entity("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_get_entity_404(self):
        # Test a random unexpected failure
        self.iam_stubber.add_client_error("get_user", 404, "NoSuchEntity")
        self.iam_stubber.activate()
        ret = yield self.actor._get_entity("test")
        self.assertIsNone(ret)
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_get_entity(self):
        # Create some valid test user objects...
        self.iam_stubber.add_response(
            # API Call
            "get_user",
            # Response
            {
                "User": {
                    "Path": "/",
                    "CreateDate": "2016-04-05T22:15:24Z",
                    "UserName": "test",
                    "UserId": "AIDAXXCXXXXXXXXXAAC2E",
                    "Arn": "arn:aws:iam::123123123123:base/test",
                }
            },
            # Call Params
            {"UserName": "test"},
        )
        self.iam_stubber.activate()
        ret = yield self.actor._get_entity("test")
        self.assertEqual(ret["UserName"], "test")

    @testing.gen_test
    def test_ensure_entity(self):
        create_mock = mock.MagicMock(name="_create_entity")
        delete_mock = mock.MagicMock(name="_delete_entity")
        get_mock = mock.MagicMock(name="_get_entity")
        self.actor._create_entity = create_mock
        self.actor._delete_entity = delete_mock
        self.actor._get_entity = get_mock
        self.actor._create_entity.side_effect = [tornado_value(None)]
        self.actor._delete_entity.side_effect = [tornado_value(None)]
        self.actor._get_entity.side_effect = [tornado_value(None)]

        # Mock out that the entity doesn't exist, and we're creating it
        self.actor._get_entity.side_effect = [tornado_value(None)]
        yield self.actor._ensure_entity("test", "present")
        create_mock.assert_called_with("test")
        self.assertFalse(delete_mock.called)
        create_mock.reset_mock()
        delete_mock.reset_mock()

        # Pretend like the entity already exists..
        entity = {
            "get_base_response": {"get_base_result": {"base": {"arn": "fake_arn"}}}
        }
        self.actor._get_entity.side_effect = [tornado_value(entity)]

        # Now if we ask to create the entity, make sure we don't make those
        # calls since the entity already exists
        self.actor._get_entity.side_effect = [tornado_value(entity)]
        yield self.actor._ensure_entity("test", "present")
        self.assertFalse(create_mock.called)
        self.assertFalse(delete_mock.called)
        create_mock.reset_mock()
        delete_mock.reset_mock()

        # Since the entity is there, lets test deleting them.. do they get
        # deleted?
        self.actor._get_entity.side_effect = [tornado_value(entity)]
        yield self.actor._ensure_entity("test", "absent")
        self.assertFalse(create_mock.called)
        delete_mock.assert_called_with("test")
        create_mock.reset_mock()
        delete_mock.reset_mock()

        # If the entity doesn't exist, make sure we don't try to delete them
        self.actor._get_entity.side_effect = [tornado_value(None)]
        yield self.actor._ensure_entity("test", "absent")
        self.assertFalse(create_mock.called)
        self.assertFalse(delete_mock.called)
        create_mock.reset_mock()
        delete_mock.reset_mock()

    @testing.gen_test
    def test_delete_entity(self):
        # stub out the get_entity_policies to return one policy name. This ensures that we first have to delete the
        # policies from the entity, then tne entity.
        self.iam_stubber.add_response(
            # API Call
            "list_user_policies",
            # Response
            {"PolicyNames": ["test"]},
            # Call Params
            {"UserName": "test"},
        )
        # Stub out getting the actual policy contents
        self.iam_stubber.add_response(
            # API Call
            "get_user_policy",
            # Response
            {
                "UserName": "test",
                "PolicyName": "test",
                "PolicyDocument": str(json.dumps({})),
            },
            # Call Params
            {"UserName": "test", "PolicyName": "test"},
        )

        # Now stub out the delete call to that particular policy
        self.iam_stubber.add_response(
            # API Call
            "delete_user_policy",
            # Response
            {},
            # Call Params
            {"UserName": "test", "PolicyName": "test"},
        )

        # Finally stub out the delete call to the user
        self.iam_stubber.add_response(
            # API Call
            "delete_user",
            # Response
            {},
            # Call Params
            {"UserName": "test"},
        )
        # Pretend it worked...
        self.iam_stubber.activate()
        yield self.actor._delete_entity("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_delete_entity_already_deleted(self):
        # Exception raised? Handle it!
        self.actor._get_entity_policies = mock.MagicMock()
        self.actor._get_entity_policies.side_effect = [tornado_value([])]
        self.iam_stubber.add_client_error("delete_user", 400, "NoSuchEntity")
        self.iam_stubber.activate()
        yield self.actor._delete_entity("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_delete_entity_other_exception(self):
        # Exception raised? Handle it!
        self.actor._get_entity_policies = mock.MagicMock()
        self.actor._get_entity_policies.side_effect = [tornado_value([])]
        self.iam_stubber.add_client_error("delete_user", 500, "Yikes!")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._delete_entity("test")

    @testing.gen_test
    def test_delete_entity_dry(self):
        # Make sure we did not call the delete function!
        self.actor._dry = True
        self.iam_stubber.activate()
        yield self.actor._delete_entity("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_create_entity(self):
        self.iam_stubber.add_response(
            # API Call
            "create_user",
            # Response,
            {
                "User": {
                    "Arn": "arn:.................",
                    "Path": "/",
                    "UserName": "test",
                    "UserId": "AQ..............C...",
                    "CreateDate": datetime(2015, 1, 1),
                }
            },
            # Call Params
            {"UserName": "test"},
        )
        self.iam_stubber.activate()
        yield self.actor._create_entity("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_create_entity_already_exists(self):
        self.iam_stubber.add_client_error("create_user", 409, "EntityAlreadyExists")
        self.iam_stubber.activate()
        yield self.actor._create_entity("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_create_entity_other_exception(self):
        self.iam_stubber.add_client_error("create_user", 500, "ServerError")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._create_entity("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_create_entity_dry(self):
        # Make sure we did not call the create function!
        self.actor._dry = True
        self.iam_stubber.activate()
        yield self.actor._create_entity("test")
        self.iam_stubber.assert_no_pending_responses()


class TestUser(testing.AsyncTestCase):
    def setUp(self):
        super(TestUser, self).setUp()
        settings.AWS_ACCESS_KEY_ID = "unit-test"
        settings.AWS_SECRET_ACCESS_KEY = "unit-test"
        settings.AWS_SESSION_TOKEN = "unit-test"
        importlib.reload(iam)

        # Create our actor object with some basics... then mock out the IAM
        # connections..
        self.actor = iam.User(
            "Unit Test",
            {
                "name": "test",
                "state": "present",
                "inline_policies": "examples/aws.iam.user/s3_example.json",
                "groups": "foo",
            },
        )

        self.iam_stubber = Stubber(self.actor.iam_conn)

    @testing.gen_test
    def test_ensure_groups(self):
        # Stub out a fake list of groups that the user is already attached to
        self.iam_stubber.add_response(
            # API Call
            "list_groups_for_user",
            # Response
            {
                "Groups": [
                    {
                        "Path": "/",
                        "GroupName": "test-group-1",
                        "GroupId": "................",
                        "Arn": "....................",
                        "CreateDate": datetime(2015, 1, 1),
                    },
                ]
            },
            # Call Params
            {"UserName": "test"},
        )
        self.iam_stubber.add_response(
            # API Call
            "add_user_to_group",
            # Response
            {},
            # Call Params
            {"UserName": "test", "GroupName": "ng1"},
        )

        # Now stub out the calls that are delete calls...
        self.iam_stubber.add_response(
            # API Call
            "remove_user_from_group",
            # Response
            {},
            # Call Params
            {"UserName": "test", "GroupName": "test-group-1"},
        )

        self.iam_stubber.activate()
        yield self.actor._ensure_groups("test", "ng1")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_groups_with_not_yet_created_user(self):
        # Create mocks for the add/remove user group methods
        self.actor._add_user_to_group = mock.MagicMock()
        self.actor._add_user_to_group.side_effect = [tornado_value(None)]
        self.actor._remove_user_from_group = mock.MagicMock()
        self.actor._remove_user_from_group.side_effect = [tornado_value(None)]

        # The user doesn't exist? No problem.. we'll move forward anyways and
        # assume we're in a dry run and the user hasn't been created, and thus
        # there are no groups.
        self.iam_stubber.add_client_error("list_groups_for_user", 404, "NoSuchEntity")
        self.iam_stubber.activate()
        yield self.actor._ensure_groups("test", ["ng1", "ng2"])
        self.actor._add_user_to_group.assert_has_calls(
            [mock.call("test", "ng1"), mock.call("test", "ng2")], any_order=True
        )
        self.assertFalse(self.actor._remove_user_from_group.called)

    @testing.gen_test
    def test_ensure_groups_with_500(self):
        self.iam_stubber.add_client_error("list_groups_for_user", 500, "Server Error")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._ensure_groups("test", ["ng1", "ng2"])
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_add_user_to_group(self):
        self.iam_stubber.add_response(
            # API Call
            "add_user_to_group",
            # Response
            {},
            # Call Params
            {"GroupName": "group", "UserName": "test"},
        )
        self.iam_stubber.activate()
        yield self.actor._add_user_to_group("test", "group")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_add_user_to_group_dry(self):
        self.iam_stubber.activate()
        self.actor._dry = True
        yield self.actor._add_user_to_group("test", "group")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_add_user_to_group_500(self):
        self.iam_stubber.add_client_error("add_user_to_group", 500, "Yikes!")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._add_user_to_group("test", "group")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_remove_user_from_group(self):
        self.iam_stubber.add_response(
            # API Call
            "remove_user_from_group",
            # Response
            {},
            # Call Params
            {"UserName": "test", "GroupName": "group"},
        )
        self.iam_stubber.activate()
        yield self.actor._remove_user_from_group("test", "group")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_remove_user_from_group_dry(self):
        self.iam_stubber.activate()
        self.actor._dry = True
        yield self.actor._remove_user_from_group("test", "group")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_remove_user_from_group_500(self):
        self.iam_stubber.add_client_error("remove_user_from_group", 500, "Yikes!")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._remove_user_from_group("test", "group")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_execute_absent(self):
        self.actor._options["state"] = "absent"
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)

    @testing.gen_test
    def test_execute_present(self):
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        self.actor._ensure_inline_policies = mock.MagicMock()
        self.actor._ensure_inline_policies.side_effect = [tornado_value(None)]
        self.actor._ensure_groups = mock.MagicMock()
        self.actor._ensure_groups.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)
        self.assertTrue(self.actor._ensure_inline_policies.called)
        self.assertTrue(self.actor._ensure_groups.called)

    @testing.gen_test
    def test_execute_present_no_policies_or_groups(self):
        self.actor._options["inline_policies"] = None
        self.actor._options["groups"] = None

        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]

        self.actor._ensure_inline_policies = mock.MagicMock()
        self.actor._ensure_groups = mock.MagicMock()

        yield self.actor._execute()

        self.assertTrue(self.actor._ensure_entity.called)
        self.assertFalse(self.actor._ensure_inline_policies.called)
        self.assertFalse(self.actor._ensure_groups.called)


class TestGroup(testing.AsyncTestCase):
    def setUp(self):
        super(TestGroup, self).setUp()
        settings.AWS_ACCESS_KEY_ID = "unit-test"
        settings.AWS_SECRET_ACCESS_KEY = "unit-test"
        settings.AWS_SESSION_TOKEN = "unit-test"
        importlib.reload(iam)

        # Create our actor object with some basics... then mock out the IAM
        # connections..
        self.actor = iam.Group(
            "Unit Test",
            {
                "name": "test",
                "state": "present",
                "inline_policies": "examples/aws.iam.user/s3_example.json",
            },
        )

        self.iam_stubber = Stubber(self.actor.iam_conn)

    @testing.gen_test
    def test_get_group_users(self):
        self.iam_stubber.add_response(
            # API Call
            "get_group",
            # Response
            {
                "Group": {
                    "GroupName": "test",
                    "Path": "/",
                    "Arn": ".........................",
                    "GroupId": ".................",
                    "CreateDate": datetime(2015, 1, 1),
                },
                "Users": [
                    {
                        "UserName": "user1",
                        "UserId": ".............................",
                        "Path": "/",
                        "Arn": ".............................",
                        "CreateDate": datetime(2015, 1, 1),
                    },
                    {
                        "UserName": "user2",
                        "UserId": ".............................",
                        "Path": "/",
                        "Arn": ".............................",
                        "CreateDate": datetime(2015, 1, 1),
                    },
                ],
            },
            # Request
            {"GroupName": "test"},
        )
        self.iam_stubber.activate()
        ret = yield self.actor._get_group_users("test")
        self.assertEqual(ret, ["user1", "user2"])
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_get_group_users_500(self):
        self.iam_stubber.add_client_error("get_group", 500, "Yikes")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._get_group_users("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_get_group_users_no_users(self):
        self.iam_stubber.add_response(
            # API Call
            "get_group",
            # Response
            {
                "Group": {
                    "GroupName": "test",
                    "Path": "/",
                    "Arn": ".........................",
                    "GroupId": ".................",
                    "CreateDate": datetime(2015, 1, 1),
                },
                "Users": [],
            },
            # Request
            {"GroupName": "test"},
        )
        self.iam_stubber.activate()
        ret = yield self.actor._get_group_users("test")
        self.assertEqual(ret, [])
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_purge_group_users_false(self):
        users = ["user1", "user2"]
        self.actor._get_group_users = mock.MagicMock()
        self.actor._get_group_users.side_effect = [tornado_value(users)]

        self.actor._remove_user_from_group = mock.MagicMock()
        self.actor._remove_user_from_group.side_effect = [
            tornado_value(None),
            tornado_value(None),
        ]

        yield self.actor._purge_group_users("test", False)

        self.assertFalse(self.actor._remove_user_from_group.called)

    @testing.gen_test
    def test_purge_group_users_true(self):
        users = ["user1", "user2"]
        self.actor._get_group_users = mock.MagicMock()
        self.actor._get_group_users.side_effect = [tornado_value(users)]

        self.actor._remove_user_from_group = mock.MagicMock()
        self.actor._remove_user_from_group.side_effect = [
            tornado_value(None),
            tornado_value(None),
        ]

        yield self.actor._purge_group_users("test", True)

        self.actor._remove_user_from_group.assert_has_calls(
            [mock.call("user1", "test"), mock.call("user2", "test")]
        )

    @testing.gen_test
    def test_execute_absent(self):
        self.actor._options["state"] = "absent"
        self.actor._purge_group_users = mock.MagicMock()
        self.actor._purge_group_users.side_effect = [tornado_value(None)]
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)
        self.assertTrue(self.actor._purge_group_users.called)

    @testing.gen_test
    def test_execute_present(self):
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_inline_policies = mock.MagicMock()
        self.actor._purge_group_users = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        self.actor._ensure_inline_policies.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)
        self.assertFalse(self.actor._purge_group_users.called)

    @testing.gen_test
    def test_execute_present_no_policies_or_groups(self):
        self.actor._options["inline_policies"] = None
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_inline_policies = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        self.actor._ensure_inline_policies.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)
        self.assertFalse(self.actor._ensure_inline_policies.called)


class TestRole(testing.AsyncTestCase):
    def setUp(self):
        super(TestRole, self).setUp()
        settings.AWS_ACCESS_KEY_ID = "unit-test"
        settings.AWS_SECRET_ACCESS_KEY = "unit-test"
        settings.AWS_SESSION_TOKEN = "unit-test"
        importlib.reload(iam)

        # Create our actor object with some basics... then mock out the IAM
        # connections..
        self.actor = iam.Role(
            "Unit Test",
            {
                "name": "test",
                "state": "present",
                "assume_role_policy_document": "examples/aws.iam.role/lambda.json",
                "inline_policies": "examples/aws.iam.user/s3_example.json",
            },
        )

        self.iam_stubber = Stubber(self.actor.iam_conn)

    @testing.gen_test
    def test_ensure_assume_role_doc_no_entity(self):
        fake_entity = None
        self.actor._get_entity = mock.MagicMock()
        self.actor._get_entity.side_effect = [tornado_value(fake_entity)]

        yield self.actor._ensure_assume_role_doc("test")

    @testing.gen_test
    def test_ensure_assume_role_doc_matches(self):
        # This is the desired doc...
        self.actor.assume_role_policy_doc = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"Service": "lambda.amazonaws.com"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }

        self.iam_stubber.add_response(
            # API Call
            "get_role",
            # Response
            {
                "Role": {
                    "Arn": "....................................",
                    "Path": "/",
                    "RoleId": ".........................",
                    "RoleName": "test",
                    "CreateDate": datetime(2019, 2, 19, 21, 3, 20),
                    "AssumeRolePolicyDocument": json.dumps(
                        self.actor.assume_role_policy_doc
                    ),
                }
            },
            # Call Params
            {"RoleName": "test"},
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_assume_role_doc("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_assume_role_doc_mismatch(self):
        # This is the desired doc...
        self.actor.assume_role_policy_doc = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"Service": "ec2.amazonaws.com"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }

        self.iam_stubber.add_response(
            # API Call
            "get_role",
            # Response
            {
                "Role": {
                    "Arn": "....................................",
                    "Path": "/",
                    "RoleId": ".........................",
                    "RoleName": "test",
                    "CreateDate": datetime(2019, 2, 19, 21, 3, 20),
                    "AssumeRolePolicyDocument": "{}",
                }
            },
            # Call Params
            {"RoleName": "test"},
        )

        self.iam_stubber.add_response(
            # API Call
            "update_assume_role_policy",
            # Response
            {},
            # Call Params
            {
                "RoleName": "test",
                "PolicyDocument": json.dumps(self.actor.assume_role_policy_doc),
            },
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_assume_role_doc("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_assume_role_doc_mismatch_dry(self):
        self.actor._dry = True
        self.actor.assume_role_policy_doc = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"Service": "ec2.amazonaws.com"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }
        self.iam_stubber.add_response(
            # API Call
            "get_role",
            # Response
            {
                "Role": {
                    "Arn": "....................................",
                    "Path": "/",
                    "RoleId": ".........................",
                    "RoleName": "test",
                    "CreateDate": datetime(2019, 2, 19, 21, 3, 20),
                    "AssumeRolePolicyDocument": "{}",
                }
            },
            # Call Params
            {"RoleName": "test"},
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_assume_role_doc("test")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_execute_absent(self):
        self.actor._options["state"] = "absent"
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        self.actor._ensure_assume_role_doc = mock.MagicMock()
        self.actor._ensure_assume_role_doc.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)
        self.assertFalse(self.actor._ensure_assume_role_doc.called)

    @testing.gen_test
    def test_execute_no_policy(self):
        self.actor._options["assume_role_policy_document"] = None
        self.actor._options["inline_policies"] = None
        self.iam_stubber.activate()
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        self.actor._ensure_assume_role_doc = mock.MagicMock()
        self.actor._ensure_assume_role_doc.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)
        self.assertTrue(self.actor._ensure_assume_role_doc.called)
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_execute(self):
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        self.actor._ensure_inline_policies = mock.MagicMock()
        self.actor._ensure_inline_policies.side_effect = [tornado_value(None)]
        self.actor._ensure_assume_role_doc = mock.MagicMock()
        self.actor._ensure_assume_role_doc.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)
        self.assertTrue(self.actor._ensure_inline_policies.called)
        self.assertTrue(self.actor._ensure_assume_role_doc.called)

    @testing.gen_test
    def test_create_entity(self):
        self.actor.assume_role_policy_doc = "{}"
        self.actor._dry = False
        self.iam_stubber.add_response(
            # API Call
            "create_role",
            # Response,
            {
                "Role": {
                    "Arn": "arn:.................",
                    "Path": "/",
                    "RoleName": "test",
                    "RoleId": "AQ..............C...",
                    "CreateDate": datetime(2015, 1, 1),
                }
            },
            # Call Params
            {"RoleName": "test", "AssumeRolePolicyDocument": '"{}"'},
        )
        self.iam_stubber.activate()
        yield self.actor._create_entity("test")
        self.iam_stubber.assert_no_pending_responses()


class TestInstanceProfile(testing.AsyncTestCase):
    def setUp(self):
        super(TestInstanceProfile, self).setUp()
        settings.AWS_ACCESS_KEY_ID = "unit-test"
        settings.AWS_SECRET_ACCESS_KEY = "unit-test"
        settings.AWS_SESSION_TOKEN = "unit-test"
        importlib.reload(iam)

        # Create our actor object with some basics... then mock out the IAM
        # connections..
        self.actor = iam.InstanceProfile(
            "Unit Test", {"name": "test", "state": "present", "role": "test"}
        )

        self.iam_stubber = Stubber(self.actor.iam_conn)

    @testing.gen_test
    def test_add_role(self):
        self.iam_stubber.add_response(
            # API Call
            "add_role_to_instance_profile",
            # Response
            {},
            # Call Params
            {"InstanceProfileName": "test", "RoleName": "testrole"},
        )
        self.iam_stubber.activate()
        yield self.actor._add_role("test", "testrole")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_add_role_409(self):
        self.iam_stubber.add_client_error(
            "add_role_to_instance_profile", 409, "NoSuchEntity"
        )
        self.iam_stubber.activate()
        yield self.actor._add_role("test", "testrole")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_add_role_500(self):
        self.iam_stubber.add_client_error("add_role_to_instance_profile", 500, "Yikes")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._add_role("test", "testrole")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_add_role_dry(self):
        self.actor._dry = True
        self.iam_stubber.activate()
        yield self.actor._add_role("test", "testrole")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_remove_role(self):
        self.iam_stubber.add_response(
            # API Call
            "remove_role_from_instance_profile",
            # Response
            {},
            # Call Params
            {"InstanceProfileName": "test", "RoleName": "testrole"},
        )
        self.iam_stubber.activate()
        yield self.actor._remove_role("test", "testrole")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_remove_role_404(self):
        self.iam_stubber.add_client_error(
            "remove_role_from_instance_profile", 404, "NoSuchEntity"
        )
        self.iam_stubber.activate()
        yield self.actor._remove_role("test", "testrole")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_remove_role_500(self):
        self.iam_stubber.add_client_error(
            "remove_role_from_instance_profile", 500, "Yikes"
        )
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._remove_role("test", "testrole")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_remove_role_dry(self):
        self.actor._dry = True
        self.iam_stubber.activate()
        yield self.actor._remove_role("test", "testrole")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_role_matching(self):
        self.iam_stubber.add_response(
            # API Call
            "get_instance_profile",
            # Response
            {
                "InstanceProfile": {
                    "Arn": "arn:aws:iam::...:instance-profile/...",
                    "CreateDate": datetime(2016, 9, 28, 19, 23, 4),
                    "InstanceProfileId": "...................",
                    "InstanceProfileName": "test",
                    "Path": "/",
                    "Roles": [
                        {
                            "Arn": "......................................",
                            "AssumeRolePolicyDocument": "{}",
                            "CreateDate": datetime(2016, 9, 28, 19, 23, 4),
                            "Path": "/",
                            "RoleId": ".....................",
                            "RoleName": "test-role",
                        }
                    ],
                    "Tags": [],
                },
            },
            # Call Params
            {"InstanceProfileName": "test"},
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_role("test", "test-role")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_role_with_no_role_set_and_is_missing_correctly(self):
        self.iam_stubber.add_response(
            # API Call
            "get_instance_profile",
            # Response
            {
                "InstanceProfile": {
                    "Arn": "arn:aws:iam::...:instance-profile/...",
                    "CreateDate": datetime(2016, 9, 28, 19, 23, 4),
                    "InstanceProfileId": "...................",
                    "InstanceProfileName": "test",
                    "Path": "/",
                    "Roles": [],
                    "Tags": [],
                },
            },
            # Call Params
            {"InstanceProfileName": "test"},
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_role("test", None)
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_role_existing_role_but_want_none(self):
        self.iam_stubber.add_response(
            # API Call
            "get_instance_profile",
            # Response
            {
                "InstanceProfile": {
                    "Arn": "arn:aws:iam::...:instance-profile/...",
                    "CreateDate": datetime(2016, 9, 28, 19, 23, 4),
                    "InstanceProfileId": "...................",
                    "InstanceProfileName": "test",
                    "Path": "/",
                    "Roles": [
                        {
                            "Arn": "......................................",
                            "AssumeRolePolicyDocument": "{}",
                            "CreateDate": datetime(2016, 9, 28, 19, 23, 4),
                            "Path": "/",
                            "RoleId": ".....................",
                            "RoleName": "wrong-role",
                        }
                    ],
                    "Tags": [],
                },
            },
            # Call Params
            {"InstanceProfileName": "test"},
        )

        self.iam_stubber.add_response(
            # API Call
            "remove_role_from_instance_profile",
            # Response
            {},
            # Call Params
            {"InstanceProfileName": "test", "RoleName": "wrong-role"},
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_role("test", None)
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_role_not_matching(self):
        self.iam_stubber.add_response(
            # API Call
            "get_instance_profile",
            # Response
            {
                "InstanceProfile": {
                    "Arn": "arn:aws:iam::...:instance-profile/...",
                    "CreateDate": datetime(2016, 9, 28, 19, 23, 4),
                    "InstanceProfileId": "...................",
                    "InstanceProfileName": "test",
                    "Path": "/",
                    "Roles": [
                        {
                            "Arn": "......................................",
                            "AssumeRolePolicyDocument": "{}",
                            "CreateDate": datetime(2016, 9, 28, 19, 23, 4),
                            "Path": "/",
                            "RoleId": ".....................",
                            "RoleName": "wrong-role",
                        }
                    ],
                    "Tags": [],
                },
            },
            # Call Params
            {"InstanceProfileName": "test"},
        )

        self.iam_stubber.add_response(
            # API Call
            "remove_role_from_instance_profile",
            # Response
            {},
            # Call Params
            {"InstanceProfileName": "test", "RoleName": "wrong-role"},
        )

        self.iam_stubber.add_response(
            # API Call
            "add_role_to_instance_profile",
            # Response
            {},
            # Call Params
            {"InstanceProfileName": "test", "RoleName": "test-role"},
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_role("test", "test-role")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_role_matching_404(self):
        self.iam_stubber.add_client_error("get_instance_profile", 404, "NoSuchEntity")
        self.iam_stubber.add_response(
            # API Call
            "add_role_to_instance_profile",
            # Response
            {},
            # Call parms
            {"InstanceProfileName": "test", "RoleName": "test-role"},
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_role("test", "test-role")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_role_matching_500(self):
        self.iam_stubber.add_client_error("get_instance_profile", 500, "Yikes")
        self.iam_stubber.activate()
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._ensure_role("test", "test-role")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_ensure_role_matching_key_error(self):
        self.iam_stubber.add_response(
            # API Call
            "get_instance_profile",
            # Response
            {
                "InstanceProfile": {
                    "Arn": "arn:aws:iam::...:instance-profile/...",
                    "CreateDate": datetime(2016, 9, 28, 19, 23, 4),
                    "InstanceProfileId": "...................",
                    "InstanceProfileName": "test",
                    "Path": "/",
                    "Roles": [],
                    "Tags": [],
                },
            },
            # Call Params
            {"InstanceProfileName": "test"},
        )

        self.iam_stubber.add_response(
            # API Call
            "add_role_to_instance_profile",
            # Response
            {},
            # Call parms
            {"InstanceProfileName": "test", "RoleName": "test-role"},
        )
        self.iam_stubber.activate()
        yield self.actor._ensure_role("test", "test-role")
        self.iam_stubber.assert_no_pending_responses()

    @testing.gen_test
    def test_execute_absent(self):
        self.actor._options["state"] = "absent"
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)

    @testing.gen_test
    def test_execute(self):
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_role = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        self.actor._ensure_role.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)
        self.assertTrue(self.actor._ensure_role.called)

    @testing.gen_test
    def test_execute_no_role(self):
        self.actor._options["role"] = None
        self.actor._ensure_entity = mock.MagicMock()
        self.actor._ensure_entity.side_effect = [tornado_value(None)]
        self.actor._ensure_role = mock.MagicMock()
        self.actor._ensure_role.side_effect = [tornado_value(None)]
        yield self.actor._execute()
        self.assertTrue(self.actor._ensure_entity.called)
        self.assertFalse(self.actor._ensure_role.called)



import logging

from tornado import httpclient
from tornado import testing
import mock

from kingpin import exceptions as kingpin_exceptions
from kingpin.actors import exceptions
from kingpin.actors import misc
from kingpin.actors.test.helper import mock_tornado
import importlib

log = logging.getLogger(__name__)


class TestNote(testing.AsyncTestCase):
    @testing.gen_test
    def test_log(self):
        note = misc.Note("Test", {"message": "Hello World"})
        note.log = mock.Mock()
        yield note._execute()
        self.assertEqual(note.log.info.call_count, 1)


class TestMacro(testing.AsyncTestCase):
    def setUp(self):
        super(TestMacro, self).setUp()
        importlib.reload(misc)

    def test_init(self):
        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_macro = mock.Mock(return_value="unit-test-macro")

        with mock.patch("kingpin.utils.convert_script_to_dict") as j2d, mock.patch(
            "kingpin.schema.validate"
        ) as schema_validate, mock.patch("kingpin.actors.utils.get_actor") as get_actor:

            j2d.return_value = {
                "desc": "unit test",
                "actor": "unit test",
                "options": {},
            }

            actor = misc.Macro(
                "Unit Test", {"macro": "test.json", "tokens": {}}, init_tokens={}
            )

            j2d.assert_called_with(script_file="unit-test-macro", tokens={})
            self.assertEqual(schema_validate.call_count, 1)
            self.assertEqual(actor.initial_actor, get_actor())

    def test_init_nested_tokens(self):
        # Quick test to ensure that a series of nested groups and macro actors
        # will pass supplied tokens all the way down to their sub actors.

        # Generate a single outer actor. This actor will create many internal
        # actors.
        init_tokens = {"SLEEP": 0}
        actor = misc.Macro(
            options={"macro": "examples/misc.macro/outer_group.yaml"},
            init_tokens=init_tokens,
        )

        # Ensure that the initial misc.Macro actor, and the initial actor it
        # created (from outer_group.yaml) both have the appropriate init
        # tokens. This helps ensure that we used .copy() properly a well as the
        # fact that the tokens were passed down appropriately.
        self.assertEqual(actor._init_tokens, init_tokens)
        self.assertEqual(actor.initial_actor._init_tokens, init_tokens)

        # Ensure that the nested misc.Macro actor from outer_macro.yaml got
        # init_tokens, AND the 'FOO' token from outer_group.yaml's own
        # definition.
        self.assertEqual(
            actor.initial_actor._actions[0]._init_tokens, {"SLEEP": 0, "FOO": "weee"}
        )

        # Next ensure that the mostly nested examples/misc.macro/inner.yaml
        # actor got the SLEEP, FOO, and DESC tokens.
        self.assertEqual(
            actor.initial_actor._actions[0].initial_actor._init_tokens,
            {"SLEEP": 0, "FOO": "weee", "DESC": "Sleeping for a while"},
        )

        # Finally, ensure the super nested
        s = actor.initial_actor._actions[0].initial_actor.initial_actor
        self.assertEqual(
            s._init_tokens, {"SLEEP": 0, "FOO": "weee", "DESC": "Sleeping for a while"}
        )

    def test_init_group(self):
        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_macro = mock.Mock(return_value="unit-test-macro")

        with mock.patch("kingpin.utils.convert_script_to_dict") as j2d, mock.patch(
            "kingpin.schema.validate"
        ) as schema_validate, mock.patch("kingpin.actors.group.Sync") as sync_actor:

            j2d.return_value = [
                {"desc": "unit test", "actor": "unit test", "options": {}}
            ]

            actor = misc.Macro("Unit Test", {"macro": "test.json"})

            j2d.assert_called_with(script_file="unit-test-macro", tokens={})
            self.assertEqual(schema_validate.call_count, 1)
            self.assertEqual(actor.initial_actor, sync_actor())

    def test_init_remote(self):
        misc.Macro._get_config_from_script = mock.Mock()
        misc.Macro._get_config_from_script.return_value = {}
        misc.Macro._check_schema = mock.Mock()
        with mock.patch("kingpin.actors.utils.get_actor"):
            with mock.patch.object(httpclient.HTTPClient, "fetch") as fetch:
                fetch.return_value.body = "foo"
                misc.Macro("Unit Test", {"macro": "http://test.json", "tokens": {}})

    def test_init_dry(self):
        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_config_from_script = mock.Mock()
        misc.Macro._get_config_from_script.return_value = {}
        misc.Macro._check_schema = mock.Mock()

        with mock.patch("kingpin.utils.convert_script_to_dict") as j2d, mock.patch(
            "kingpin.schema.validate"
        ), mock.patch("kingpin.actors.utils.get_actor"):
            j2d.return_value = {
                "desc": "unit test",
                "actor": "unit test",
                "options": {},
            }

            actor = misc.Macro(
                "Unit Test",
                {"macro": "examples/test/sleep.json", "tokens": {}},
                dry=True,
            )

            self.assertTrue(actor.initial_actor._dry)

    def test_init_with_errors(self):

        # Remote files are prohibited for now
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro("Unit Test", {"macro": "ftp://fail.test.json", "tokens": {}})

        # Remote file with bad URL
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro("Unit Test", {"macro": "http://fail.test.json", "tokens": {}})

        # Non-existent file
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro("Unit Test", {"macro": "dontcreatethis.json", "tokens": {}})

        # We don't want the rest of the tests failing on downloading this file.
        misc.Macro._get_macro = mock.Mock(return_value="unit-test-file")

        # Schema failure
        with mock.patch("kingpin.utils.convert_script_to_dict") as j2d:
            j2d.return_value = {
                "desc": "unit test",
                "options": {},  # `actor` keyword is missing
            }

            with self.assertRaises(exceptions.UnrecoverableActorFailure):
                misc.Macro("Unit Test", {"macro": "test.json", "tokens": {}})

        # JSON syntax error
        with mock.patch("kingpin.utils.convert_script_to_dict") as j2d:

            j2d.side_effect = kingpin_exceptions.InvalidScript("Fail!")

            with self.assertRaises(exceptions.UnrecoverableActorFailure):
                misc.Macro("Unit Test", {"macro": "test.json", "tokens": {}})

    @testing.gen_test
    def test_execute(self):

        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_macro = mock.Mock()
        misc.Macro._get_config_from_script = mock.Mock()
        misc.Macro._get_config_from_script.return_value = {}
        misc.Macro._check_schema = mock.Mock()

        with mock.patch("kingpin.actors.utils.get_actor") as get_actor:
            actor = misc.Macro(
                "Unit Test", {"macro": "test.json", "tokens": {}}, dry=True
            )

            get_actor().execute = mock_tornado()
            yield actor._execute()

            self.assertEqual(get_actor().execute._call_count, 1)

    @testing.gen_test
    def test_orgchart(self):

        misc.Macro._get_macro = mock.Mock(name="unittestmacro")
        misc.Macro._get_config_from_script = mock.Mock(
            return_value=[{"actor": "misc.Sleep", "options": {"sleep": 0}}]
        )
        actor = misc.Macro("Unit test", {"macro": "test"})

        self.assertEqual(len(actor.get_orgchart()), 3)  # Macro, Group, Sleep
        self.assertEqual(type(actor.get_orgchart()[0]), dict)


class TestSleep(testing.AsyncTestCase):
    @testing.gen_test
    def test_execute(self):
        # Call the executor and test it out
        actor = misc.Sleep("Unit Test Action", {"sleep": 0.1})
        res = yield actor.execute()

        # Make sure we fired off an alert.
        self.assertEqual(res, None)

    @testing.gen_test
    def test_execute_with_str(self):
        # Call the executor and test it out
        actor = misc.Sleep("Unit Test Action", {"sleep": "0.5"})
        res = yield actor.execute()

        # Make sure we fired off an alert.
        self.assertEqual(res, None)


class TestGenericHTTP(testing.AsyncTestCase):
    @testing.gen_test
    def test_execute_dry(self):
        actor = misc.GenericHTTP(
            "Unit Test Action", {"url": "http://example.com"}, dry=True
        )

        actor._fetch = mock_tornado()

        yield actor.execute()

        self.assertEqual(actor._fetch._call_count, 0)

    @testing.gen_test
    def test_execute(self):
        actor = misc.GenericHTTP("Unit Test Action", {"url": "http://example.com"})
        actor._fetch = mock_tornado({"success": {"code": 200}})

        yield actor.execute()

    @testing.gen_test
    def test_execute_data_json(self):
        actor = misc.GenericHTTP(
            "Unit Test Action",
            {"url": "http://example.com", "data-json": {"foo": "bar"}},
        )
        actor._fetch = mock_tornado({"success": {"code": 200}})

        yield actor.execute()

    @testing.gen_test
    def test_execute_fail(self):
        actor = misc.GenericHTTP("Unit Test Action", {"url": "http://example.com"})
        error = httpclient.HTTPError(code=401, response={})
        actor._fetch = mock_tornado(exc=error)

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()



class KingpinException(Exception):
    """Base Exception"""


class InvalidScript(KingpinException):
    """Raised when an invalid script schema was detected"""


class InvalidScriptName(KingpinException):
    """Raised when the script name does not end on .yaml or .json"""



#!/usr/bin/env python
"""CLI Script Runner for Kingpin."""

import argparse
import json
import logging
import os
import sys

from tornado import gen
from tornado import ioloop

from kingpin import utils
from kingpin.actors import utils as actor_utils
from kingpin.actors import exceptions as actor_exceptions
from kingpin.actors.misc import Macro
from kingpin.version import __version__


log = logging.getLogger(__name__)


# We handle all the exceptions ourselves, so additional log statements from
# BOTO are not needed.
logging.getLogger("boto").setLevel(logging.CRITICAL)
logging.getLogger("botocore").setLevel(logging.CRITICAL)

# Initial option handler to set up the basic application environment.
parser = argparse.ArgumentParser(description="Kingpin v%s" % __version__)
parser.set_defaults(verbose=True)

# Job Configuration
parser.add_argument(
    "-j",
    "--json",
    "-s",
    "--script",
    dest="script",
    help="Path to JSON/YAML Deployment Script",
)
parser.add_argument(
    "-a",
    "--actor",
    dest="actor",
    help="Name of an Actor to execute (overrides --script)",
)
parser.add_argument(
    "-E",
    "--explain",
    dest="explain",
    action="store_true",
    help="Explain how an actor works. Requires --actor.",
    default=False,
)
parser.add_argument(
    "-p",
    "--param",
    dest="params",
    action="append",
    help="Actor Parameter to set (ie, warn_on_failure=true)",
    default=[],
)
parser.add_argument(
    "-o",
    "--option",
    dest="options",
    action="append",
    help="Actor Options to set (ie, elb_name=foobar)",
    default=[],
)
parser.add_argument(
    "-d", "--dry", dest="dry", action="store_true", help="Executes a dry run only."
)
parser.add_argument(
    "--build-only",
    dest="build_only",
    action="store_true",
    help="Compile the input JSON without executing any runs",
)
parser.add_argument(
    "--orgchart",
    dest="orgchart",
    help="Save the orgchart into file. Requires --build-only",
)

# Logging Configuration
parser.add_argument(
    "-l",
    "--level",
    dest="level",
    default="info",
    help="Set logging level (INFO|WARN|DEBUG|ERROR)",
)
parser.add_argument(
    "-D",
    "--debug",
    dest="level_debug",
    default=False,
    action="store_true",
    help="Equivalent to --level=DEBUG",
)
parser.add_argument(
    "-c",
    "--color",
    dest="color",
    default=False,
    action="store_true",
    help="Colorize the log output",
)

args = parser.parse_args()


def kingpin_fail(message: str):
    parser.print_help()
    sys.stderr.write("\nError: %s\n" % message)
    sys.exit(1)


def get_main_actor(dry: bool):
    env_tokens = dict(os.environ)

    # Cannot specify a script file an an actor at the same time.
    if args.script and args.actor:
        kingpin_fail("You may only specify --actor or --script, not both!")

    if args.actor:
        ActorClass = actor_utils.get_actor_class(args.actor)
        parameters = dict([i.split("=") for i in args.params])
        options = dict([i.split("=") for i in args.options])

        return ActorClass(
            options=options, dry=dry, init_tokens=env_tokens, **parameters
        )

    # Actor not specified. Process JSON file.
    try:
        script = args.script or sys.argv[1] if sys.argv else None
    except Exception as e:
        kingpin_fail(
            "%s You must specify --script or provide it as first argument." % e
        )

    return Macro(
        desc="Kingpin", options={"macro": script, "tokens": env_tokens}, dry=dry
    )


@gen.coroutine
def main():

    if args.actor and args.explain:
        ActorClass = actor_utils.get_actor_class(args.actor)
        print((ActorClass.__doc__))
        sys.exit(0)

    if args.build_only:
        try:
            actor = get_main_actor(dry=False)
        except Exception as e:
            log.critical(e)
            sys.exit(1)

        if args.orgchart:
            log.info("Creating organizational chart into %s" % args.orgchart)
            try:
                orgdata = actor.get_orgchart()
            except Exception as e:
                log.critical(e)
                sys.exit(2)

            with open(args.orgchart, "w") as output:
                output.write(json.dumps(orgdata))

        sys.exit(0)

    if not args.dry:
        # Lets maybe do a dry run first anyways...
        skip_dry = False
        try:
            skip_dry = utils.str2bool(os.getenv("SKIP_DRY", "False"), strict=True)
        except ValueError:
            log.warning("SKIP_DRY is not a valid boolean-like. Defaulting to False.")
            pass

        if skip_dry:
            # Okay, the user really does not want to do a dry run... fine
            log.warning("")
            log.warning("*** You have disabled the pre-check dry run.")
            log.warning("*** Execution will begin with no expectation of success.")
            log.warning("")
        else:
            log.info("Rehearsing... Break a leg!")

            try:
                dry_actor = get_main_actor(dry=True)
                yield dry_actor.execute()
            except actor_exceptions.ActorException as e:
                log.critical("Dry run failed. Reason:")
                log.critical(e)
                sys.exit(2)

            log.info("Rehearsal OK! Performing!")

    # Begin doing real stuff!
    try:
        runner = get_main_actor(dry=args.dry)

        log.info("")
        log.warning("Lights, camera ... action!")
        log.info("")
        yield runner.execute()
    except actor_exceptions.ActorException as e:
        log.error("Kingpin encountered mistakes during the play.")
        log.error(e)
        sys.exit(2)


def begin():
    # Set up logging before we do anything else
    if args.level_debug:
        args.level = "DEBUG"
    utils.setup_root_logger(level=args.level, color=args.color)

    try:
        ioloop.IOLoop.instance().run_sync(main)
    except KeyboardInterrupt:
        log.info("CTRL-C Caught, shutting down")
        sys.exit(130)  # Standard KeyboardInterrupt exit code.
    except Exception:
        # Skip traceback that involves tornado's libraries.
        import traceback

        trace_lines = traceback.format_exc().splitlines()
        skip_next = False
        for l in trace_lines:
            if "tornado" in l:
                skip_next = True
                continue
            if not skip_next:
                print(l)
            skip_next = False
        sys.exit(3)


if __name__ == "__main__":  # pragma: no cover
    begin()



import importlib
import os
import sys
import unittest
from unittest import mock
from kingpin.actors.exceptions import ActorException
from kingpin.actors.misc import Macro, Sleep


class TestDeploy(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestDeploy, self).__init__(*args, **kwargs)

        self.kingpin_bin_deploy = None

    def setUp(self):
        """
        Loading the kingpin.bin.deploy module will call argparse.ArgumentParser.parse_args() which
        needs specific arguments to be passed in sys.argv. We need to save the original sys.argv here
        and restore it in tearDown. Furthermore, we will then patch sys.argv to have the correct
        values during import of the kingpin.bin.deploy module.
        """
        self.original_argv = sys.argv

    def tearDown(self):
        if self.kingpin_bin_deploy:
            del self.kingpin_bin_deploy
        sys.argv = self.original_argv

    def _import_kingpin_bin_deploy(self):
        if self.kingpin_bin_deploy:
            del self.kingpin_bin_deploy
        self.kingpin_bin_deploy = importlib.import_module("kingpin.bin.deploy")
        importlib.reload(self.kingpin_bin_deploy)

    ############################################################################
    #  kingpin_fail
    ############################################################################

    @mock.patch("sys.argv", ["kingpin"])
    def test_kingpin_fail(self):
        self._import_kingpin_bin_deploy()
        with self.assertRaises(SystemExit) as cm:
            self.kingpin_bin_deploy.kingpin_fail("testing")
            self.assertEqual(cm.exception.code, 1)

    ############################################################################
    #  get_main_actor
    ############################################################################

    @mock.patch("sys.argv", ["kingpin", "--script", "examples/test/sleep.yaml"])
    def test_get_main_actor_with_script(self):
        self._import_kingpin_bin_deploy()
        result = self.kingpin_bin_deploy.get_main_actor(True)
        self.assertEqual(str(type(result)), str(Macro))
        self.assertEqual(result.option("macro"), "examples/test/sleep.yaml")

    @mock.patch(
        "sys.argv", ["kingpin", "--actor", "misc.Sleep", "--option", "sleep=0.1"]
    )
    def test_get_main_actor_with_actor(self):
        self._import_kingpin_bin_deploy()
        result = self.kingpin_bin_deploy.get_main_actor(True)
        self.assertEqual(str(type(result)), str(Sleep))
        self.assertEqual(result.option("sleep"), "0.1")

    @mock.patch(
        "sys.argv",
        ["kingpin", "--script", "examples/test/sleep.yaml", "--actor", "misc.Sleep"],
    )
    def test_get_main_actor_with_both_script_and_actor(self):
        self._import_kingpin_bin_deploy()
        with self.assertRaises(SystemExit) as cm:
            self.kingpin_bin_deploy.get_main_actor(True)
        self.assertEqual(cm.exception.code, 1)

    @mock.patch("sys.argv", ["kingpin"])
    def test_get_main_actor_with_neither_script_nor_actor(self):
        self._import_kingpin_bin_deploy()
        with self.assertRaises(SystemExit) as cm:
            self.kingpin_bin_deploy.get_main_actor(True)
        self.assertEqual(cm.exception.code, 1)

    ############################################################################
    #  main
    ############################################################################

    @mock.patch("sys.argv", ["kingpin"])
    def test_main_with_good_runner(self):
        self._import_kingpin_bin_deploy()
        with mock.patch("kingpin.bin.deploy.get_main_actor") as mock_get_main_actor:
            mock_get_main_actor.return_value = Sleep(options={"sleep": 0.1}, dry=True)
            self.kingpin_bin_deploy.main()
            mock_get_main_actor.assert_called()

    @mock.patch("sys.argv", ["kingpin"])
    @mock.patch.dict(os.environ, {"SKIP_DRY": "True"})
    def test_main_with_skip_dry(self):
        self._import_kingpin_bin_deploy()
        with mock.patch("kingpin.bin.deploy.get_main_actor") as mock_get_main_actor:
            mock_get_main_actor.return_value = Sleep(options={"sleep": 0.1}, dry=True)
            self.kingpin_bin_deploy.main()
            mock_get_main_actor.assert_called()

    @mock.patch("sys.argv", ["kingpin"])
    @mock.patch.dict(os.environ, {"SKIP_DRY": "not-a-boolean-like-string"})
    def test_main_with_skip_dry_invalid(self):
        self._import_kingpin_bin_deploy()
        with mock.patch("kingpin.bin.deploy.get_main_actor") as mock_get_main_actor:
            mock_get_main_actor.return_value = Sleep(options={"sleep": 0.1}, dry=True)
            self.kingpin_bin_deploy.main()
            mock_get_main_actor.assert_called()

    @mock.patch("sys.argv", ["kingpin"])
    def test_main_with_bad_runner(self):
        self._import_kingpin_bin_deploy()
        with mock.patch("kingpin.bin.deploy.get_main_actor") as mock_get_main_actor:
            mock_get_main_actor.side_effect = ActorException("testing")
            with self.assertRaises(SystemExit) as cm:
                self.kingpin_bin_deploy.main()
                self.assertEqual(cm.exception.code, 2)
            mock_get_main_actor.assert_called()

    @mock.patch("sys.argv", ["kingpin", "--dry"])
    def test_main_with_bad_runner_dry(self):
        self._import_kingpin_bin_deploy()
        with mock.patch("kingpin.bin.deploy.get_main_actor") as mock_get_main_actor:
            mock_get_main_actor.side_effect = ActorException("testing")
            with self.assertRaises(SystemExit) as cm:
                self.kingpin_bin_deploy.main()
                self.assertEqual(cm.exception.code, 2)
            mock_get_main_actor.assert_called()

    @mock.patch("sys.argv", ["kingpin", "--actor", "misc.Sleep", "--explain"])
    def test_main_with_actor_and_explain(self):
        self._import_kingpin_bin_deploy()
        with self.assertRaises(SystemExit) as cm:
            self.kingpin_bin_deploy.main()
            self.assertEqual(cm.exception.code, 0)

    @mock.patch(
        "sys.argv",
        ["kingpin", "--build-only", "--actor", "misc.Sleep", "--option", "sleep=0.1"],
    )
    def test_main_with_build_only_good_actor(self):
        self._import_kingpin_bin_deploy()
        with self.assertRaises(SystemExit) as cm:
            self.kingpin_bin_deploy.main()
            self.assertEqual(cm.exception.code, 0)

    @mock.patch(
        "sys.argv",
        ["kingpin", "--build-only", "--actor", "somefakeactor"],
    )
    def test_main_with_build_only_bad_actor(self):
        self._import_kingpin_bin_deploy()
        with self.assertRaises(SystemExit) as cm:
            self.kingpin_bin_deploy.main()
            self.assertEqual(cm.exception.code, 1)

    @mock.patch(
        "sys.argv",
        [
            "kingpin",
            "--build-only",
            "--orgchart",
            "orgchart.tmp~",
            "--actor",
            "misc.Sleep",
            "--option",
            "sleep=0.1",
        ],
    )
    def test_main_with_build_only_and_org_chart(self):
        self._import_kingpin_bin_deploy()
        with self.assertRaises(SystemExit) as cm:
            self.kingpin_bin_deploy.main()
            self.assertEqual(cm.exception.code, 0)

    @mock.patch(
        "sys.argv",
        [
            "kingpin",
            "--build-only",
            "--orgchart",
            "orgchart.tmp~",
            "--actor",
            "misc.Sleep",
            "--option",
            "sleep=0.1",
        ],
    )
    def test_main_with_build_only_and_bad_org_chart(self):
        self._import_kingpin_bin_deploy()
        actor = Sleep(options={"sleep": 0.1}, dry=True)
        actor.get_orgchart = mock.Mock(side_effect=Exception("testing"))
        with mock.patch("kingpin.bin.deploy.get_main_actor") as mock_get_main_actor:
            mock_get_main_actor.return_value = actor
            with self.assertRaises(SystemExit) as cm:
                self.kingpin_bin_deploy.main()
                self.assertEqual(cm.exception.code, 2)

    ############################################################################
    #  begin
    ############################################################################

    @mock.patch(
        "sys.argv", ["kingpin", "--actor", "misc.Sleep", "--option", "sleep=0.1"]
    )
    def test_begin(self):
        self._import_kingpin_bin_deploy()
        with mock.patch("kingpin.bin.deploy.main"):
            with self.assertRaises(SystemExit) as cm:
                self.kingpin_bin_deploy.begin()
                self.assertEqual(cm.exception.code, 0)

    @mock.patch(
        "sys.argv",
        ["kingpin", "--actor", "misc.Sleep", "--option", "sleep=0.1", "--debug"],
    )
    def test_begin_with_debug(self):
        self._import_kingpin_bin_deploy()
        with mock.patch("kingpin.bin.deploy.main"):
            with self.assertRaises(SystemExit) as cm:
                self.kingpin_bin_deploy.begin()
                self.assertEqual(cm.exception.code, 0)
                self.assertEqual(self.kingpin_bin_deploy.log.level, 10)

    @mock.patch(
        "sys.argv", ["kingpin", "--actor", "misc.Sleep", "--option", "sleep=0.1"]
    )
    def test_begin_keyboard_interrupt(self):
        self._import_kingpin_bin_deploy()
        with mock.patch("tornado.ioloop.IOLoop.instance") as mock_ioloop_instance:
            mock_ioloop_instance.side_effect = KeyboardInterrupt()
            with self.assertRaises(SystemExit) as cm:
                self.kingpin_bin_deploy.begin()
                self.assertEqual(cm.exception.code, 130)



import jsonschema

from kingpin.actors import exceptions


class REQUIRED(object):
    """Meta class to identify required arguments for actors."""


class StringCompareBase(object):
    """Meta class to identify the desired state for a resource.

    This basic type of constant allows someone to easily define a set of valid
    strings for their option and have the base actor class automatically
    validate the inputs against those strings.
    """

    valid = None

    @classmethod
    def validate(self, option):
        if option not in self.valid:
            raise exceptions.InvalidOptions(
                "%s not valid, use: %s" % (option, self.valid)
            )


class STATE(StringCompareBase):
    """Meta class to identify the desired state for a resource.

    Simple tester for 'present' or 'absent' on actors. Used for any actor thats
    idempotent and used to ensure some state of a resource.
    """

    valid = ("present", "absent")


class SchemaCompareBase(object):
    """Meta class that compares the schema of a dict against rules."""

    SCHEMA = None

    @classmethod
    def validate(self, option):
        try:
            jsonschema.Draft4Validator(self.SCHEMA).validate(option)
        except jsonschema.exceptions.ValidationError as e:
            raise exceptions.InvalidOptions(
                "Supplied parameter does not match schema: %s" % e
            )



import json
import os

import unittest

from kingpin import exceptions
from kingpin import schema


class TestSchema(unittest.TestCase):
    def setUp(self, *args, **kwargs):
        super(TestSchema, self).setUp(*args, **kwargs)

        dirname, filename = os.path.split(os.path.abspath(__file__))
        self.examples = "%s/../../examples" % dirname

    def test_validate_with_simple_json(self):
        j = json.loads(open("%s/simple.json" % self.examples).read())
        ret = schema.validate(j)
        self.assertEqual(None, ret)

    def test_validate_with_complex_json(self):
        j = json.loads(open("%s/complex.json" % self.examples).read())
        ret = schema.validate(j)
        self.assertEqual(None, ret)

    def test_validate_with_invalid_json(self):
        j = {"this": "is", "invalid": "ok"}
        with self.assertRaises(exceptions.InvalidScript):
            schema.validate(j)

    def test_validate_with_array_syntax(self):
        j = [{"actor": "some actor"}]
        schema.validate(j)

    def test_validate_with_invalid_array(self):
        j = [{"garbage": "json"}]
        with self.assertRaises(exceptions.InvalidScript):
            schema.validate(j)



"""Simple integration tests for the AWS IAM actors."""

from nose.plugins.attrib import attr
import logging

from tornado import testing

from kingpin.actors.aws import iam

__author__ = "Matt Wise (matt@nextdoor.com)"

log = logging.getLogger(__name__)
logging.getLogger("boto").setLevel(logging.INFO)


class IntegrationIAMUsers(testing.AsyncTestCase):

    integration = True

    name = "kingpin-integration-test"
    inline_policies = ["examples/aws.iam.user/s3_example.json"]
    region = "us-east-1"

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_01_ensure_user_absent(self):
        actor = iam.User("Test", {"name": self.name, "state": "absent"}, dry=False)
        yield actor.execute()

    @attr("aws", "integration", "dry")
    @testing.gen_test(timeout=60)
    def integration_02a_create_user_dry(self):
        actor = iam.User(
            "Test",
            {
                "name": self.name,
                "state": "present",
                "inline_policies": self.inline_policies,
                "inline_policies_purge": True,
            },
            dry=True,
        )

        yield actor.execute()

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_02b_create_user(self):
        actor = iam.User(
            "Test",
            {
                "name": self.name,
                "inline_policies": self.inline_policies,
                "inline_policies_purge": True,
            },
            dry=False,
        )

        yield actor.execute()

    # Final cleanup -- delete our test user!
    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_09_ensure_user_absent(self):
        actor = iam.User("Test", {"name": self.name, "state": "absent"}, dry=False)
        yield actor.execute()


class IntegrationIAMGroups(testing.AsyncTestCase):

    integration = True

    name = "kingpin-integration-test"
    inline_policies = ["examples/aws.iam.user/s3_example.json"]
    region = "us-east-1"

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_01_ensure_group_absent(self):
        actor = iam.Group("Test", {"name": self.name, "state": "absent"}, dry=False)
        yield actor.execute()

    @attr("aws", "integration", "dry")
    @testing.gen_test(timeout=60)
    def integration_02a_create_group_dry(self):
        actor = iam.Group(
            "Test",
            {
                "name": self.name,
                "state": "present",
                "inline_policies": self.inline_policies,
                "inline_policies_purge": True,
            },
            dry=True,
        )

        yield actor.execute()

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_02b_create_group(self):
        actor = iam.Group(
            "Test",
            {
                "name": self.name,
                "inline_policies": self.inline_policies,
                "inline_policies_purge": True,
            },
            dry=False,
        )

        yield actor.execute()

    # Final cleanup -- delete our test group!
    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_09_ensure_group_absent(self):
        actor = iam.Group("Test", {"name": self.name, "state": "absent"}, dry=False)
        yield actor.execute()


class IntegrationIAMRoles(testing.AsyncTestCase):

    integration = True

    name = "kingpin-integration-test"
    inline_policies = ["examples/aws.iam.user/s3_example.json"]
    region = "us-east-1"

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_01_ensure_role_absent(self):
        actor = iam.Role("Test", {"name": self.name, "state": "absent"}, dry=False)
        yield actor.execute()

    @attr("aws", "integration", "dry")
    @testing.gen_test(timeout=60)
    def integration_02a_create_role_dry(self):
        actor = iam.Role(
            "Test",
            {
                "name": self.name,
                "state": "present",
                "inline_policies": self.inline_policies,
                "inline_policies_purge": True,
            },
            dry=True,
        )

        yield actor.execute()

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_02b_create_role(self):
        actor = iam.Role(
            "Test",
            {
                "name": self.name,
                "inline_policies": self.inline_policies,
                "inline_policies_purge": True,
            },
            dry=False,
        )

        yield actor.execute()

    # Final cleanup -- delete our test role!
    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_09_ensure_role_absent(self):
        actor = iam.Role("Test", {"name": self.name, "state": "absent"}, dry=False)
        yield actor.execute()


class IntegrationIAMInstanceProfiles(testing.AsyncTestCase):

    integration = True

    name = "kingpin-integration-test"
    inline_policies = ["examples/aws.iam.user/s3_example.json"]
    region = "us-east-1"

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_01_ensure_profile_absent(self):
        actor = iam.InstanceProfile(
            "Test", {"name": self.name, "state": "absent"}, dry=False
        )
        yield actor.execute()

    @attr("aws", "integration", "dry")
    @testing.gen_test(timeout=60)
    def integration_02a_create_profile_dry(self):
        actor = iam.InstanceProfile(
            "Test",
            {
                "name": self.name,
                "state": "present",
                "inline_policies": self.inline_policies,
                "inline_policies_purge": True,
            },
            dry=True,
        )

        yield actor.execute()

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_02b_create_profile(self):
        actor = iam.InstanceProfile(
            "Test",
            {
                "name": self.name,
                "inline_policies": self.inline_policies,
                "inline_policies_purge": True,
            },
            dry=False,
        )

        yield actor.execute()

    # Final cleanup -- delete our test profile!
    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_09_ensure_profile_absent(self):
        actor = iam.InstanceProfile(
            "Test", {"name": self.name, "state": "absent"}, dry=False
        )
        yield actor.execute()



"""
:mod:`kingpin.actors.aws.settings`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Common settings used by many of the `kingpin.actors.aws` modules.
"""

import os

# By default, this means that Boto will make HTTP calls at instantiation time to
# determine whether or not credentials are available from the metadata service.
#
# During tests, we mock these out to blank strings to prevent these calls.
AWS_ACCESS_KEY_ID = os.getenv("AWS_ACCESS_KEY_ID", None)
AWS_SECRET_ACCESS_KEY = os.getenv("AWS_SECRET_ACCESS_KEY", None)
AWS_SESSION_TOKEN = os.getenv("AWS_SESSION_TOKEN", None)

# kingpin is pretty fast which can leads to API throttling. You can set you own
# boto3 configuration by using the standard AWS env vars, but in the absence of
# them, we try to set you some sane defaults based on the boto3 documentation
# and our experience with running kingpin as scale.
#
# Docs: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html
AWS_MAX_ATTEMPTS = int(os.getenv("AWS_MAX_ATTEMPTS", 10))
AWS_RETRY_MODE = os.getenv("AWS_RETRY_MODE", "standard")

# Set to "" (an empty string) to disable.
KINGPIN_CFN_HASH_OUTPUT_KEY = os.getenv("KINGPIN_CFN_HASH_OUTPUT_KEY", "KingpinCfnHash")

# Instead of specifying the role_arn in each CloudFormation actor, you can set a
# default role.
KINGPIN_CFN_DEFAULT_ROLE_ARN = os.getenv("KINGPIN_CFN_DEFAULT_ROLE_ARN", None)
